
#include "wanpipe_sppp.h"



/*
 * 
 */
const struct cp chap = {
	PPP_CHAP, IDX_CHAP, CP_AUTH, "chap",
	wpsppp_null, wpsppp_null, wpsppp_chap_open, wpsppp_chap_close,
	wpsppp_chap_TO, 0, 0, 0,
	wpsppp_chap_tlu, wpsppp_chap_tld, wpsppp_null, wpsppp_null,
	wpsppp_chap_scr
};



/*
 *--------------------------------------------------------------------------*
 *                                                                          *
 *                        The CHAP implementation.                          *
 *                                                                          *
 *--------------------------------------------------------------------------*
 */

/*
 * The authentication protocols don't employ a full-fledged state machine as
 * the control protocols do, since they do have Open and Close events, but
 * not Up and Down, nor are they explicitly terminated.  Also, use of the
 * authentication protocols may be different in both directions (this makes
 * sense, think of a machine that never accepts incoming calls but only
 * calls out, it doesn't require the called party to authenticate itself).
 *
 * Our state machine for the local authentication protocol (we are requesting
 * the peer to authenticate) looks like:
 *
 *						    RCA-
 *	      +--------------------------------------------+
 *	      V					    scn,tld|
 *	  +--------+			       Close   +---------+ RCA+
 *	  |	   |<----------------------------------|	 |------+
 *   +--->| Closed |				TO*    | Opened	 | sca	|
 *   |	  |	   |-----+		       +-------|	 |<-----+
 *   |	  +--------+ irc |		       |       +---------+
 *   |	    ^		 |		       |	   ^
 *   |	    |		 |		       |	   |
 *   |	    |		 |		       |	   |
 *   |	 TO-|		 |		       |	   |
 *   |	    |tld  TO+	 V		       |	   |
 *   |	    |	+------->+		       |	   |
 *   |	    |	|	 |		       |	   |
 *   |	  +--------+	 V		       |	   |
 *   |	  |	   |<----+<--------------------+	   |
 *   |	  | Req-   | scr				   |
 *   |	  | Sent   |					   |
 *   |	  |	   |					   |
 *   |	  +--------+					   |
 *   | RCA- |	| RCA+					   |
 *   +------+	+------------------------------------------+
 *   scn,tld	  sca,irc,ict,tlu
 *
 *
 *   with:
 *
 *	Open:	LCP reached authentication phase
 *	Close:	LCP reached terminate phase
 *
 *	RCA+:	received reply (pap-req, chap-response), acceptable
 *	RCN:	received reply (pap-req, chap-response), not acceptable
 *	TO+:	timeout with restart counter >= 0
 *	TO-:	timeout with restart counter < 0
 *	TO*:	reschedule timeout for CHAP
 *
 *	scr:	send request packet (none for PAP, chap-challenge)
 *	sca:	send ack packet (pap-ack, chap-success)
 *	scn:	send nak packet (pap-nak, chap-failure)
 *	ict:	initialize re-challenge timer (CHAP only)
 *
 *	tlu:	this-layer-up, LCP reaches network phase
 *	tld:	this-layer-down, LCP enters terminate phase
 *
 * Note that in CHAP mode, after sending a new challenge, while the state
 * automaton falls back into Req-Sent state, it doesn't signal a tld
 * event to LCP, so LCP remains in network phase.  Only after not getting
 * any response (or after getting an unacceptable response), CHAP closes,
 * causing LCP to enter terminate phase.
 *
 * With PAP, there is no initial request that can be sent.  The peer is
 * expected to send one based on the successful negotiation of PAP as
 * the authentication protocol during the LCP option negotiation.
 *
 * Incoming authentication protocol requests (remote requests
 * authentication, we are peer) don't employ a state machine at all,
 * they are simply answered.  Some peers [Ascend P50 firmware rev
 * 4.50] react allergically when sending IPCP requests while they are
 * still in authentication phase (thereby violating the standard that
 * demands that these NCP packets are to be discarded), so we keep
 * track of the peer demanding us to authenticate, and only proceed to
 * phase network once we've seen a positive acknowledge for the
 * authentication.
 */


/*
 * Handle incoming CHAP packets.
 */
void
wpsppp_chap_input(struct sppp *sp, void *skb)
{
	struct lcp_header *h;
	int len;//, x;
	unsigned char *value, digest[AUTHKEYLEN], dsize;
	unsigned char *name;
	int value_len, name_len;
	struct wp_MD5Context ctx;
	
	//strcpy(sp->hisauth.name, "RYELLE_LINX");
	//strcpy(sp->hisauth.secret, "ryelle");

	len = wpabs_skb_len(skb);
	
	if (len < 4) {
		if (debug){
			wpabs_debug_event( "%s: chap invalid packet length: %d bytes\n",
				sp->name, len);
		}
		return;
	}

	DBG_CHDLC("%s(): line: %d\n", __FUNCTION__, __LINE__);

	h = (struct lcp_header *)wpabs_skb_data(skb);
		
	if (len > wpabs_ntohs (h->len))
		len = wpabs_ntohs (h->len);

	switch (h->type) {
	/* challenge, failure and success are his authproto */
	case CHAP_CHALLENGE:
	
		value = 1 + (unsigned char*)(h+1);
		value_len = value[-1];
		name = value + value_len;
		name_len = len - value_len - 5;
		
		DBG_CHDLC("%s(): line: %d\n", __FUNCTION__, __LINE__);
		sp->chap_stats.rx_challenge++;

		if (name_len < 0) {
			sp->chap_stats.rx_bad_resp++;

			if (debug){
				wpabs_debug_event( "%s: chap corrupted challenge <%s id=0x%x len=%d",
					sp->name,
					wpsppp_auth_type_name(PPP_CHAP, h->type),
					h->ident, wpabs_ntohs(h->len));
				
				wpsppp_print_bytes((unsigned char*) (h+1), len-4);
				wpabs_debug_event(">\n");
			}
			break;
		}

		if (debug){
			wpabs_debug_event(
			    	"%s: chap input <%s id=0x%x len=%d name=",
				sp->name,
			    	wpsppp_auth_type_name(PPP_CHAP, h->type),
			    	h->ident,
			    	wpabs_ntohs(h->len));
			
			wpsppp_print_string((char*) name, name_len);
			
			wpabs_debug_event(" value-size=%d value=", value_len);
			wpsppp_print_bytes(value, value_len);
			wpabs_debug_event(">\n");
		}
		
		/* Compute reply value. */
		wp_MD5Init(&ctx);
		wp_MD5Update(&ctx, &h->ident, 1);
		wp_MD5Update(&ctx, sp->myauth.secret, wpsppp_strnlen(sp->myauth.secret, AUTHKEYLEN));
		wp_MD5Update(&ctx, value, value_len);
		wp_MD5Final(digest, &ctx);
		dsize = sizeof digest;

		{
		unsigned char data_buf [MAX_AUTHENTICATE_PACKET_LEN];
		int dlen=0;	
		
		wpabs_memcpy(&data_buf[dlen],(char *)&dsize,sizeof dsize);
		dlen+=sizeof dsize;
		wpabs_memcpy(&data_buf[dlen],digest,sizeof digest);
		dlen+=sizeof digest;
		wpabs_memcpy(&data_buf[dlen],sp->myauth.name,
				(int)wpsppp_strnlen(sp->myauth.name, AUTHNAMELEN));
		dlen+=(int)wpsppp_strnlen(sp->myauth.name, AUTHNAMELEN);
		
		sp->chap_stats.tx_response++;
		wpsppp_auth_send(sp, PPP_CHAP, CHAP_RESPONSE, h->ident,
			       dlen,data_buf);
		}
		break;

	case CHAP_SUCCESS:
	
		sp->chap_stats.rx_success++;

		if (debug){
			wpabs_debug_event( "%s: chap success ", sp->name);
			if (len > 4) {
				wpabs_debug_event(": ");
				wpsppp_print_string((char*)(h + 1), len - 4);
			}
			wpabs_debug_event("\n");
		}
		sp->myauth.authenticated = 0;	
		sp->pp_flags &= ~PP_NEEDAUTH;

		/* indicate protocol 'up' */
		sp->pp_link_state=SPPP_LINK_UP;
		wpabs_debug_event( "%s: protocol up\n", sp->name);
		wpsppp_connect_indication(sp, 0);

		/* set LCP state to opened */
		sp->lcp.state = LCP_STATE_OPENED;
		wpsppp_ipcp_open (sp);
		break;

	case CHAP_FAILURE:

		sp->chap_stats.rx_failure++;

		if (debug){
			wpabs_debug_event( "%s: chap failure", sp->name);
			if (len > 4) {
				wpabs_debug_event(": ");
				wpsppp_print_string((char*)(h + 1), len - 4);
			}
			wpabs_debug_event("\n");
		}else
			wpabs_debug_event( "%s: chap failure\n", sp->name);
		
		/* await LCP shutdown by authenticator */
		break;

	/* response is my authproto */
	case CHAP_RESPONSE:

		wpabs_debug_event( "%s : CHAP_RESPONSE - not implemented yet.\n",
			sp->name);
		break;
		
		value = 1 + (unsigned char*)(h+1);
		value_len = value[-1];
		name = value + value_len;
		name_len = len - value_len - 5;
		if (name_len < 0) {
			if (debug) {
				wpabs_debug_event(
				    SPP_FMT "chap corrupted response "
				    "<%s id=0x%x len=%d",
				    SPP_ARGS(sp),
				    wpsppp_auth_type_name(PPP_CHAP, h->type),
				    h->ident, wpabs_ntohs(h->len));
				wpsppp_print_bytes((unsigned char*)(h+1), len-4);
				wpabs_debug_event(">\n");
			}
			break;
		}
		if (h->ident != sp->confid[IDX_CHAP]) {
			if (debug)
				wpabs_debug_event( SPP_FMT "chap dropping response for old ID "
				    "(got %d, expected %d)\n",
				    SPP_ARGS(sp),
				    h->ident, sp->confid[IDX_CHAP]);
			break;
		}
		if (name_len != wpsppp_strnlen(sp->hisauth.name, AUTHNAMELEN)
		    || wpabs_memcmp(name, sp->hisauth.name, name_len) != 0) {
			wpabs_debug_event( SPP_FMT "chap response, his name ",
			    	SPP_ARGS(sp));
			wpsppp_print_string((char*)name, name_len);
			wpabs_debug_event(" != expected ");
			wpsppp_print_string((char*)sp->hisauth.name,
					  wpsppp_strnlen(sp->hisauth.name, AUTHNAMELEN));
			wpabs_debug_event("\n");
		}
		if (debug) {
			wpabs_debug_event( SPP_FMT "chap input(%s) "
				"<%s id=0x%x len=%d name=",
			    	SPP_ARGS(sp),
			    	wpsppp_state_name(sp->state[IDX_CHAP]),
			    	wpsppp_auth_type_name(PPP_CHAP, h->type),
			    	h->ident, wpabs_ntohs (h->len));
			wpsppp_print_string((char*)name, name_len);
			wpabs_debug_event(" value-size=%d value=", value_len);
			wpsppp_print_bytes(value, value_len);
			wpabs_debug_event(">\n");
		}
		if (value_len != AUTHKEYLEN) {
			if (debug)
				wpabs_debug_event( SPP_FMT "chap bad hash value length: "
					"%d bytes, should be %d\n",
				    	SPP_ARGS(sp), value_len,
				    	AUTHKEYLEN);
			break;
		}

		wp_MD5Init(&ctx);
		wp_MD5Update(&ctx, &h->ident, 1);
		wp_MD5Update(&ctx, sp->hisauth.secret,
			  wpsppp_strnlen(sp->hisauth.secret, AUTHKEYLEN));
		wp_MD5Update(&ctx, sp->myauth.challenge, AUTHKEYLEN);
		wp_MD5Final(digest, &ctx);

#define FAILMSG "Failed..."
#define SUCCMSG "Welcome!"

		if (value_len != sizeof digest ||
		    wpabs_memcmp(digest, value, value_len) != 0) {
			// action scn, tld
			{
			unsigned char data_buf [MAX_AUTHENTICATE_PACKET_LEN];
			int dlen=0;	
			wpabs_memcpy(&data_buf[dlen],(unsigned char *)FAILMSG,sizeof(FAILMSG) - 1);
			dlen+=sizeof(FAILMSG) - 1;
	
			wpsppp_auth_send(sp, PPP_CHAP, CHAP_FAILURE, h->ident,
				       dlen,data_buf);
			}
			chap.tld(sp);
			break;
		}
		// action sca, perhaps tlu
		if (sp->state[IDX_CHAP] == STATE_REQ_SENT ||
		    sp->state[IDX_CHAP] == STATE_OPENED){
			
			{
			unsigned char data_buf [MAX_AUTHENTICATE_PACKET_LEN];
			int dlen=0;	
			wpabs_memcpy(&data_buf[dlen],(unsigned char *)SUCCMSG,sizeof(SUCCMSG) - 1);
			dlen+=sizeof(SUCCMSG) - 1;

			wpsppp_auth_send(sp, PPP_CHAP, CHAP_SUCCESS, h->ident,
				       dlen,data_buf);
			}
		}
		
		if (sp->state[IDX_CHAP] == STATE_REQ_SENT) {
			wpsppp_cp_change_state(&chap, sp, STATE_OPENED);
			chap.tlu(sp);
		}
		break;

	default:
		sp->chap_stats.rx_unknown++;

		wpabs_debug_event( "Unknown CHAP packet type !!\n");
		
		/* Unknown CHAP packet type -- ignore. */
		
		if (sp->pp_flags & PP_DEBUG){
			wpabs_debug_event( "%s: chap unknown input <type=0x%x id=0x%xh len=%d",
				sp->name,
			    	h->type, h->ident, wpabs_ntohs(h->len));
			
			wpsppp_print_bytes((unsigned char*)(h+1), len-4);

			wpabs_debug_event(">\n");
		}
		
		break;

	}
}

/*
 * Send a PAP or CHAP proto packet.
 *
 * Varadic function, each of the elements for the ellipsis is of type
 * ``size_t mlen, const unsigned char *msg''.  Processing will stop iff
 * mlen == 0.
 * NOTE: never declare variadic functions with types subject to type
 * promotion (i.e. unsigned char). This is asking for big trouble depending
 * on the architecture you are on...
 */

void
wpsppp_auth_send(struct sppp *sp, unsigned int protocol, unsigned int type,
	       unsigned int id, int len, unsigned char *data_buf)
{
#if 0
	if (sp->pp_flags & PP_DEBUG){
		wpabs_debug_event( "%s: %s output <%s id=0x%x len=%d",
			sp->name,
		    	(protocol == PPP_PAP ? "pap" : "chap"),
		    	wpsppp_auth_type_name(protocol, type),
		    	id, len);
		
		wpsppp_print_bytes((unsigned char*) (data_buf), len);
		wpabs_debug_event(">\n");
	}
#endif
	wpsppp_cp_send (sp, protocol, type, id, len, data_buf);
}


void
wpsppp_chap_init(struct sppp *sp)
{
	/* Chap doesn't have STATE_INITIAL at all. */
	sp->state[IDX_CHAP] = STATE_CLOSED;
	sp->fail_counter[IDX_CHAP] = 0;
	sp->pp_seq[IDX_CHAP] = 0;
	sp->pp_rseq[IDX_CHAP] = 0;
#if 0
//FIXME: ALEX what is this?????
#if defined(__FreeBSD__) && __FreeBSD__ >= 3
	callout_handle_init(&sp->ch[IDX_CHAP]);
#endif
#endif
}

void
wpsppp_chap_open(struct sppp *sp)
{
	wpabs_debug_event( "wpsppp_chap_open()\n");
	
	if (sp->myauth.proto == PPP_CHAP &&
	    (sp->lcp.opts & (1 << LCP_OPT_AUTH_PROTO)) != 0) {
		/* we are authenticator for CHAP, start it */
		chap.scr(sp);
		sp->rst_counter[IDX_CHAP] = sp->lcp.max_configure;
		wpsppp_cp_change_state(&chap, sp, STATE_REQ_SENT);
	}
	/* nothing to be done if we are peer, await a challenge */
}

void
wpsppp_chap_close(struct sppp *sp)
{
	wpabs_debug_event( "wpsppp_chap_close()\n");

	if (sp->state[IDX_CHAP] != STATE_CLOSED)
		wpsppp_cp_change_state(&chap, sp, STATE_CLOSED);
}

void
wpsppp_chap_TO(void *cookie)
{
	struct sppp *sp = (struct sppp *)cookie;

	wpabs_debug_event( "wpsppp_chap_TO()\n");
	

	if (debug)
		wpabs_debug_event( SPP_FMT "chap TO(%s) rst_counter = %d\n",
		    SPP_ARGS(sp),
		    wpsppp_state_name(sp->state[IDX_CHAP]),
		    sp->rst_counter[IDX_CHAP]);

	if (--sp->rst_counter[IDX_CHAP] < 0)
		/* TO- event */
		switch (sp->state[IDX_CHAP]) {
		case STATE_REQ_SENT:
			chap.tld(sp);
			wpsppp_cp_change_state(&chap, sp, STATE_CLOSED);
			break;
		}
	else
		/* TO+ (or TO*) event */
		switch (sp->state[IDX_CHAP]) {
		case STATE_OPENED:
			/* TO* event */
			sp->rst_counter[IDX_CHAP] = sp->lcp.max_configure;
			/* FALLTHROUGH */
		case STATE_REQ_SENT:
			chap.scr(sp);
			/* wpsppp_cp_change_state() will restart the timer */
			wpsppp_cp_change_state(&chap, sp, STATE_REQ_SENT);
			break;
		}

}

void
wpsppp_chap_tlu(struct sppp *sp)
{
	int i;

	wpabs_debug_event( "wpsppp_chap_tlu()\n");
	
	i = 0;
	sp->rst_counter[IDX_CHAP] = sp->lcp.max_configure;

	/*
	 * Some broken CHAP implementations (Conware CoNet, firmware
	 * 4.0.?) don't want to re-authenticate their CHAP once the
	 * initial challenge-response exchange has taken place.
	 * Provide for an option to avoid rechallenges.
	 */
	if ((sp->hisauth.flags & AUTHFLAG_NORECHALLENGE) == 0) {
		/*
		 * Compute the re-challenge timeout.  This will yield
		 * a number between 300 and 810 seconds.
		 */
		//i = 300 + ((unsigned)(random() & 0xff00) >> 7);
		i = 300;// + ((wpabs_get_systemticks() & 0x0000ff00) >> 7);
		TIMEOUT(chap.TO, (void *)sp, i);

	}

	if (debug) {
		wpabs_debug_event("chap %s, ",
			SPP_ARGS(sp),
		    	sp->pp_phase == PHASE_NETWORK? "reconfirmed": "tlu");

		if ((sp->hisauth.flags & AUTHFLAG_NORECHALLENGE) == 0)
			wpabs_debug_event( "next re-challenge in %d seconds\n", i);
		else
			wpabs_debug_event( "re-challenging supressed\n");
	}


	/* indicate to LCP that we need to be closed down */
	sp->lcp.protos |= (1 << IDX_CHAP);

	if (sp->pp_flags & PP_NEEDAUTH) {
		/*
		 * Remote is authenticator, but his auth proto didn't
		 * complete yet.  Defer the transition to network
		 * phase.
		 */
		
		return;
	}
	

	/*
	 * If we are already in phase network, we are done here.  This
	 * is the case if this is a dummy tlu event after a re-challenge.
	 */
	if (sp->pp_phase != PHASE_NETWORK)
		wpsppp_phase_network(sp);
}

void
wpsppp_chap_tld(struct sppp *sp)
{

	if (debug)
		wpabs_debug_event( SPP_FMT "chap tld\n", SPP_ARGS(sp));
	UNTIMEOUT( (void *)sp );
	sp->lcp.protos &= ~(1 << IDX_CHAP);

	//lcp.Close(sp);
	
	//????
	/* Shut down the PPP link. */
	sp->lcp.magic = wpabs_get_systemticks();
	sp->lcp.state = LCP_STATE_CLOSED;
	sp->ipcp.state = IPCP_STATE_CLOSED;
	//wpsppp_clear_timeout (sp);
	/* Initiate negotiation. */
	wpsppp_lcp_open (sp);
}

void
wpsppp_chap_scr(struct sppp *sp)
{
	unsigned long *ch, seed;
	unsigned char clen;

	wpabs_debug_event( "wpsppp_chap_scr()\n");
	
	/* Compute random challenge. */
	ch = (unsigned long *)sp->myauth.challenge;
	
#if defined(__FreeBSD__) && __FreeBSD__ >= 3
	wpabs_get_random_bytes(&seed, sizeof seed);
#else
	{
		//struct timeval tv;
		//microtime(&tv);
		//seed = tv.tv_sec ^ tv.tv_usec;
		seed = wpabs_get_systemticks();
	}
#endif
	ch[0] = wpabs_get_systemticks() & 0x000000ff;//seed ^ random();
	ch[1] = wpabs_get_systemticks() & 0x0000ff00;//seed ^ random();
	ch[2] = wpabs_get_systemticks() & 0x00ff0000;//seed ^ random();
	ch[3] = wpabs_get_systemticks() & 0xff000000;//seed ^ random();
	clen = AUTHKEYLEN;

	sp->confid[IDX_CHAP] = ++sp->pp_seq[IDX_CHAP];


	{
	unsigned char data_buf [MAX_AUTHENTICATE_PACKET_LEN];
	int dlen=0;	
	wpabs_memcpy(&data_buf[dlen],(char *)&clen,sizeof clen);
	dlen+=sizeof clen;
	
	wpabs_memcpy(&data_buf[dlen],sp->myauth.challenge,(int)AUTHKEYLEN);
	dlen+=(int)AUTHKEYLEN;
	
	wpabs_memcpy(&data_buf[dlen],sp->myauth.name,(int)wpsppp_strnlen(sp->myauth.name, AUTHNAMELEN));
	dlen+=(int)wpsppp_strnlen(sp->myauth.name, AUTHNAMELEN);
	
	wpsppp_auth_send(sp, PPP_CHAP, CHAP_CHALLENGE, sp->confid[IDX_CHAP],
		       dlen,data_buf);

	}	
}

void wpsppp_pap_scr(struct sppp *sp)
{
	unsigned char idlen, pwdlen;
	
	//sp->confid[IDX_PAP] = ++sp->pp_seq[IDX_PAP];
	pwdlen = wpsppp_strnlen(sp->myauth.secret, AUTHKEYLEN);
	idlen = wpsppp_strnlen(sp->myauth.name, AUTHNAMELEN);

	{
	unsigned char data_buf [MAX_AUTHENTICATE_PACKET_LEN];
	int dlen=0;	
	wpabs_memcpy(&data_buf[dlen],(char *)&idlen,sizeof idlen);
	dlen+=sizeof idlen;
	wpabs_memcpy(&data_buf[dlen],sp->myauth.name,(int)idlen);
	dlen+=(int)idlen;
	wpabs_memcpy(&data_buf[dlen],(char *)&pwdlen,sizeof pwdlen);
	dlen+=sizeof pwdlen;
	wpabs_memcpy(&data_buf[dlen],sp->myauth.secret,(int)pwdlen);
	dlen+=(int)pwdlen;

	sp->pap_stats.tx_auth_rqst++;
	wpsppp_auth_send(sp, PPP_PAP, PAP_REQ, sp->confid[IDX_PAP],
		       dlen,data_buf);
	}	
}


/*
 * Start MD5 accumulation.  Set bit count to 0 and buffer to mysterious
 * initialization constants.
 */
void
wp_MD5Init(struct wp_MD5Context *ctx)
{
	ctx->buf[0] = 0x67452301;
	ctx->buf[1] = 0xefcdab89;
	ctx->buf[2] = 0x98badcfe;
	ctx->buf[3] = 0x10325476;

	ctx->bytes[0] = 0;
	ctx->bytes[1] = 0;
}

/*
 * Update context to reflect the concatenation of another buffer full
 * of bytes.
 */

void
wp_MD5Update(struct wp_MD5Context *ctx, md5byte *buf, unsigned len)
{
	unsigned int t;

	/* Update byte count */

	t = ctx->bytes[0];
	if ((ctx->bytes[0] = t + len) < t)
		ctx->bytes[1]++;	/* Carry from low to high */

	t = 64 - (t & 0x3f);	/* Space available in ctx->in (at least 1) */
	if (t > len) {
		wpabs_memcpy((md5byte *)ctx->in + 64 - t, buf, len);
		return;
	}
	/* First chunk is an odd size */
	wpabs_memcpy((md5byte *)ctx->in + 64 - t, buf, t);
	byteSwap(ctx->in, 16);
	wp_MD5Transform(ctx->buf, ctx->in);
	buf += t;
	len -= t;

	/* Process data in 64-byte chunks */
	while (len >= 64) {
		wpabs_memcpy(ctx->in, buf, 64);
		byteSwap(ctx->in, 16);
		wp_MD5Transform(ctx->buf, ctx->in);
		buf += 64;
		len -= 64;
	}

	/* Handle any remaining bytes of data. */
	wpabs_memcpy(ctx->in, buf, len);
}

/*
 * Final wrapup - pad to 64-byte boundary with the bit pattern 
 * 1 0* (64-bit count of bits processed, MSB-first)
 */
void
wp_MD5Final(md5byte digest[16], struct wp_MD5Context *ctx)
{
	int count = ctx->bytes[0] & 0x3f;	/* Number of bytes in ctx->in */
	md5byte *p = (md5byte *)ctx->in + count;

	/* Set the first char of padding to 0x80.  There is always room. */
	*p++ = 0x80;

	/* Bytes of padding needed to make 56 bytes (-8..55) */
	count = 56 - 1 - count;

	if (count < 0) {	/* Padding forces an extra block */
		wpabs_memset(p, 0, count + 8);
		byteSwap(ctx->in, 16);
		wp_MD5Transform(ctx->buf, ctx->in);
		p = (md5byte *)ctx->in;
		count = 56;
	}
	wpabs_memset(p, 0, count);
	byteSwap(ctx->in, 14);

	/* Append length in bits and transform */
	ctx->in[14] = ctx->bytes[0] << 3;
	ctx->in[15] = ctx->bytes[1] << 3 | ctx->bytes[0] >> 29;
	wp_MD5Transform(ctx->buf, ctx->in);

	byteSwap(ctx->buf, 4);
	wpabs_memcpy(digest, ctx->buf, 16);
	wpabs_memset(ctx, 0, sizeof(*ctx)); /* In case it's sensitive */
}

#ifndef ASM_MD5

/* The four core functions - F1 is optimized somewhat */

/* #define F1(x, y, z) (x & y | ~x & z) */
#define F1(x, y, z) (z ^ (x & (y ^ z)))
#define F2(x, y, z) F1(z, x, y)
#define F3(x, y, z) (x ^ y ^ z)
#define F4(x, y, z) (y ^ (x | ~z))

/* This is the central step in the MD5 algorithm. */
#define wp_MD5STEP(f,w,x,y,z,in,s) \
	 (w += f(x,y,z) + in, w = (w<<s | w>>(32-s)) + x)

/*
 * The core of the MD5 algorithm, this alters an existing MD5 hash to
 * reflect the addition of 16 longwords of new data.  MD5Update blocks
 * the data and converts bytes into longwords for this routine.
 */
void
wp_MD5Transform(unsigned int buf[4], unsigned int const in[16])
{
	register unsigned int a, b, c, d;

	a = buf[0];
	b = buf[1];
	c = buf[2];
	d = buf[3];

	wp_MD5STEP(F1, a, b, c, d, in[0] + 0xd76aa478, 7);
	wp_MD5STEP(F1, d, a, b, c, in[1] + 0xe8c7b756, 12);
	wp_MD5STEP(F1, c, d, a, b, in[2] + 0x242070db, 17);
	wp_MD5STEP(F1, b, c, d, a, in[3] + 0xc1bdceee, 22);
	wp_MD5STEP(F1, a, b, c, d, in[4] + 0xf57c0faf, 7);
	wp_MD5STEP(F1, d, a, b, c, in[5] + 0x4787c62a, 12);
	wp_MD5STEP(F1, c, d, a, b, in[6] + 0xa8304613, 17);
	wp_MD5STEP(F1, b, c, d, a, in[7] + 0xfd469501, 22);
	wp_MD5STEP(F1, a, b, c, d, in[8] + 0x698098d8, 7);
	wp_MD5STEP(F1, d, a, b, c, in[9] + 0x8b44f7af, 12);
	wp_MD5STEP(F1, c, d, a, b, in[10] + 0xffff5bb1, 17);
	wp_MD5STEP(F1, b, c, d, a, in[11] + 0x895cd7be, 22);
	wp_MD5STEP(F1, a, b, c, d, in[12] + 0x6b901122, 7);
	wp_MD5STEP(F1, d, a, b, c, in[13] + 0xfd987193, 12);
	wp_MD5STEP(F1, c, d, a, b, in[14] + 0xa679438e, 17);
	wp_MD5STEP(F1, b, c, d, a, in[15] + 0x49b40821, 22);

	wp_MD5STEP(F2, a, b, c, d, in[1] + 0xf61e2562, 5);
	wp_MD5STEP(F2, d, a, b, c, in[6] + 0xc040b340, 9);
	wp_MD5STEP(F2, c, d, a, b, in[11] + 0x265e5a51, 14);
	wp_MD5STEP(F2, b, c, d, a, in[0] + 0xe9b6c7aa, 20);
	wp_MD5STEP(F2, a, b, c, d, in[5] + 0xd62f105d, 5);
	wp_MD5STEP(F2, d, a, b, c, in[10] + 0x02441453, 9);
	wp_MD5STEP(F2, c, d, a, b, in[15] + 0xd8a1e681, 14);
	wp_MD5STEP(F2, b, c, d, a, in[4] + 0xe7d3fbc8, 20);
	wp_MD5STEP(F2, a, b, c, d, in[9] + 0x21e1cde6, 5);
	wp_MD5STEP(F2, d, a, b, c, in[14] + 0xc33707d6, 9);
	wp_MD5STEP(F2, c, d, a, b, in[3] + 0xf4d50d87, 14);
	wp_MD5STEP(F2, b, c, d, a, in[8] + 0x455a14ed, 20);
	wp_MD5STEP(F2, a, b, c, d, in[13] + 0xa9e3e905, 5);
	wp_MD5STEP(F2, d, a, b, c, in[2] + 0xfcefa3f8, 9);
	wp_MD5STEP(F2, c, d, a, b, in[7] + 0x676f02d9, 14);
	wp_MD5STEP(F2, b, c, d, a, in[12] + 0x8d2a4c8a, 20);

	wp_MD5STEP(F3, a, b, c, d, in[5] + 0xfffa3942, 4);
	wp_MD5STEP(F3, d, a, b, c, in[8] + 0x8771f681, 11);
	wp_MD5STEP(F3, c, d, a, b, in[11] + 0x6d9d6122, 16);
	wp_MD5STEP(F3, b, c, d, a, in[14] + 0xfde5380c, 23);
	wp_MD5STEP(F3, a, b, c, d, in[1] + 0xa4beea44, 4);
	wp_MD5STEP(F3, d, a, b, c, in[4] + 0x4bdecfa9, 11);
	wp_MD5STEP(F3, c, d, a, b, in[7] + 0xf6bb4b60, 16);
	wp_MD5STEP(F3, b, c, d, a, in[10] + 0xbebfbc70, 23);
	wp_MD5STEP(F3, a, b, c, d, in[13] + 0x289b7ec6, 4);
	wp_MD5STEP(F3, d, a, b, c, in[0] + 0xeaa127fa, 11);
	wp_MD5STEP(F3, c, d, a, b, in[3] + 0xd4ef3085, 16);
	wp_MD5STEP(F3, b, c, d, a, in[6] + 0x04881d05, 23);
	wp_MD5STEP(F3, a, b, c, d, in[9] + 0xd9d4d039, 4);
	wp_MD5STEP(F3, d, a, b, c, in[12] + 0xe6db99e5, 11);
	wp_MD5STEP(F3, c, d, a, b, in[15] + 0x1fa27cf8, 16);
	wp_MD5STEP(F3, b, c, d, a, in[2] + 0xc4ac5665, 23);

	wp_MD5STEP(F4, a, b, c, d, in[0] + 0xf4292244, 6);
	wp_MD5STEP(F4, d, a, b, c, in[7] + 0x432aff97, 10);
	wp_MD5STEP(F4, c, d, a, b, in[14] + 0xab9423a7, 15);
	wp_MD5STEP(F4, b, c, d, a, in[5] + 0xfc93a039, 21);
	wp_MD5STEP(F4, a, b, c, d, in[12] + 0x655b59c3, 6);
	wp_MD5STEP(F4, d, a, b, c, in[3] + 0x8f0ccc92, 10);
	wp_MD5STEP(F4, c, d, a, b, in[10] + 0xffeff47d, 15);
	wp_MD5STEP(F4, b, c, d, a, in[1] + 0x85845dd1, 21);
	wp_MD5STEP(F4, a, b, c, d, in[8] + 0x6fa87e4f, 6);
	wp_MD5STEP(F4, d, a, b, c, in[15] + 0xfe2ce6e0, 10);
	wp_MD5STEP(F4, c, d, a, b, in[6] + 0xa3014314, 15);
	wp_MD5STEP(F4, b, c, d, a, in[13] + 0x4e0811a1, 21);
	wp_MD5STEP(F4, a, b, c, d, in[4] + 0xf7537e82, 6);
	wp_MD5STEP(F4, d, a, b, c, in[11] + 0xbd3af235, 10);
	wp_MD5STEP(F4, c, d, a, b, in[2] + 0x2ad7d2bb, 15);
	wp_MD5STEP(F4, b, c, d, a, in[9] + 0xeb86d391, 21);

	buf[0] += a;
	buf[1] += b;
	buf[2] += c;
	buf[3] += d;
}

#endif

